# A minimal testcase loading an IDS file and checking that the structure built is ok

import logging

from . import imaspy

root_logger = logging.getLogger("imaspy")
logger = root_logger
logger.setLevel(logging.DEBUG)


def test_load_minimal():
    ids = imaspy.ids_classes.IDSRoot(
        0, 0, xml_path="assets/IDS_minimal.xml", verbosity=2
    )  # Create a empty IDSs

    # Check if the datatypes are loaded correctly
    assert ids.minimal.a.data_type == "FLT_0D"
    assert ids.minimal.ids_properties.comment.data_type == "STR_0D"

    # Check the documentation
    # assert ids.minimal.a.documentation == "A float"
    # assert ids.minimal.ids_properties.documentation == "Properties of this IDS"
    # assert ids.minimal.ids_properties.comment.documentation == "A string comment"

    # Check the units
    # assert ids.minimal.a.units == "unitless"

    # Check the static/dynamic/constant annotation
    # assert ids.minimal.a.type == "static"
    # assert ids.minimal.ids_properties.comment.type == "constant"


def test_load_multiple_minimal():
    ids = imaspy.ids_classes.IDSRoot(
        0, 0, xml_path="assets/IDS_minimal.xml", verbosity=2
    )  # Create a empty IDSs

    # Check if the datatypes are loaded correctly
    assert ids.minimal.a.data_type == "FLT_0D"
    assert ids.minimal.ids_properties.comment.data_type == "STR_0D"

    ids2 = imaspy.ids_classes.IDSRoot(
        0, 0, xml_path="assets/IDS_minimal.xml", verbosity=2
    )  # Create a empty IDSs

    # Check if the datatypes are loaded correctly
    assert ids2.minimal.a.data_type == "FLT_0D"
    assert ids2.minimal.ids_properties.comment.data_type == "STR_0D"



# BackendArray, AbstractDataStore and AbstractWritableDataStore
# are taken from xarray c07160dd2d627a021e58515cbd7753c11fb56d94
# The following unneeded functions have been removed:
# variables
# attrs
# dimensions
#
#
# Copyright 2014-2020, xarray Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Mapping

import imaspy.backends.xarray_core_indexing as indexing
from imaspy.backends.xarray_core_utils import NdimSizeLenMixin


class BackendArray(NdimSizeLenMixin, indexing.ExplicitlyIndexed):
    __slots__ = ()

    def __array__(self, dtype=None):
        key = indexing.BasicIndexer((slice(None),) * self.ndim)
        return np.asarray(self[key], dtype=dtype)


class AbstractDataStore(Mapping):
    __slots__ = ()

    def __iter__(self):
        return iter(self.variables)

    def __getitem__(self, key):
        return self.variables[key]

    def __len__(self):
        return len(self.variables)

    def get_dimensions(self):  # pragma: no cover
        raise NotImplementedError()

    def get_attrs(self):  # pragma: no cover
        raise NotImplementedError()

    def get_variables(self):  # pragma: no cover
        raise NotImplementedError()

    def get_encoding(self):
        return {}

    def load(self):
        """
        This loads the variables and attributes simultaneously.
        A centralized loading function makes it easier to create
        data stores that do automatic encoding/decoding.

        For example::

            class SuffixAppendingDataStore(AbstractDataStore):

                def load(self):
                    variables, attributes = AbstractDataStore.load(self)
                    variables = {'%s_suffix' % k: v
                                 for k, v in variables.items()}
                    attributes = {'%s_suffix' % k: v
                                  for k, v in attributes.items()}
                    return variables, attributes

        This function will be called anytime variables or attributes
        are requested, so care should be taken to make sure its fast.
        """
        variables = FrozenDict(
            (_decode_variable_name(k), v) for k, v in self.get_variables().items()
        )
        attributes = FrozenDict(self.get_attrs())
        return variables, attributes

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()


class AbstractWritableDataStore(AbstractDataStore):
    __slots__ = ()

    def encode(self, variables, attributes):
        """
        Encode the variables and attributes in this store

        Parameters
        ----------
        variables : dict-like
            Dictionary of key/value (variable name / xr.Variable) pairs
        attributes : dict-like
            Dictionary of key/value (attribute name / attribute) pairs

        Returns
        -------
        variables : dict-like
        attributes : dict-like

        """
        variables = {k: self.encode_variable(v) for k, v in variables.items()}
        attributes = {k: self.encode_attribute(v) for k, v in attributes.items()}
        return variables, attributes

    def encode_variable(self, v):
        """encode one variable"""
        return v

    def encode_attribute(self, a):
        """encode one attribute"""
        return a

    def set_dimension(self, dim, length):  # pragma: no cover
        raise NotImplementedError()

    def set_attribute(self, k, v):  # pragma: no cover
        raise NotImplementedError()

    def set_variable(self, k, v):  # pragma: no cover
        raise NotImplementedError()

    def store_dataset(self, dataset):
        """
        in stores, variables are all variables AND coordinates
        in xarray.Dataset variables are variables NOT coordinates,
        so here we pass the whole dataset in instead of doing
        dataset.variables
        """
        self.store(dataset, dataset.attrs)

    def store(
        self,
        variables,
        attributes,
        check_encoding_set=frozenset(),
        writer=None,
        unlimited_dims=None,
    ):
        """
        Top level method for putting data on this store, this method:
          - encodes variables/attributes
          - sets dimensions
          - sets variables

        Parameters
        ----------
        variables : dict-like
            Dictionary of key/value (variable name / xr.Variable) pairs
        attributes : dict-like
            Dictionary of key/value (attribute name / attribute) pairs
        check_encoding_set : list-like
            List of variables that should be checked for invalid encoding
            values
        writer : ArrayWriter
        unlimited_dims : list-like
            List of dimension names that should be treated as unlimited
            dimensions.
        """
        if writer is None:
            writer = ArrayWriter()

        variables, attributes = self.encode(variables, attributes)

        self.set_attributes(attributes)
        self.set_dimensions(variables, unlimited_dims=unlimited_dims)
        self.set_variables(
            variables, check_encoding_set, writer, unlimited_dims=unlimited_dims
        )

    def set_attributes(self, attributes):
        """
        This provides a centralized method to set the dataset attributes on the
        data store.

        Parameters
        ----------
        attributes : dict-like
            Dictionary of key/value (attribute name / attribute) pairs
        """
        for k, v in attributes.items():
            self.set_attribute(k, v)

    def set_variables(self, variables, check_encoding_set, writer, unlimited_dims=None):
        """
        This provides a centralized method to set the variables on the data
        store.

        Parameters
        ----------
        variables : dict-like
            Dictionary of key/value (variable name / xr.Variable) pairs
        check_encoding_set : list-like
            List of variables that should be checked for invalid encoding
            values
        writer : ArrayWriter
        unlimited_dims : list-like
            List of dimension names that should be treated as unlimited
            dimensions.
        """

        for vn, v in variables.items():
            name = _encode_variable_name(vn)
            check = vn in check_encoding_set
            target, source = self.prepare_variable(
                name, v, check, unlimited_dims=unlimited_dims
            )

            writer.add(source, target)

    def set_dimensions(self, variables, unlimited_dims=None):
        """
        This provides a centralized method to set the dimensions on the data
        store.

        Parameters
        ----------
        variables : dict-like
            Dictionary of key/value (variable name / xr.Variable) pairs
        unlimited_dims : list-like
            List of dimension names that should be treated as unlimited
            dimensions.
        """
        if unlimited_dims is None:
            unlimited_dims = set()

        existing_dims = self.get_dimensions()

        dims = {}
        for v in unlimited_dims:  # put unlimited_dims first
            dims[v] = None
        for v in variables.values():
            dims.update(dict(zip(v.dims, v.shape)))

        for dim, length in dims.items():
            if dim in existing_dims and length != existing_dims[dim]:
                raise ValueError(
                    "Unable to update size for existing dimension"
                    "%r (%d != %d)" % (dim, length, existing_dims[dim])
                )
            elif dim not in existing_dims:
                is_unlimited = dim in unlimited_dims
                self.set_dimension(dim, length, is_unlimited)


class WritableIMASDataStore(AbstractWritableDataStore):
    __slots__ = ()

    def encode(self, variables, attributes):
        # Encode all varibles IMAS-style. Equivalent to xarrays CFDataStore, repeated here
        # for reference:
        # def encode(self, variables, attributes):
        #    variables = {k: self.encode_variable(v) for k, v in variables.items()}
        #    attributes = {k: self.encode_attribute(v) for k, v in attributes.items()}
        #    return variables, attributes
        variables, attributes = imas_encoder(variables, attributes)
        variables = {k: self.encode_variable(v) for k, v in variables.items()}
        attributes = {k: self.encode_attribute(v) for k, v in attributes.items()}
        return variables, attributes



# ExplicitIndexer and ExplicitlyIndexed
# are taken from xarray c07160dd2d627a021e58515cbd7753c11fb56d94
#
# This code can be modified if you explicitly note your
# changes. However, it would be better to wrap these
# classes instead
#
#
# Copyright 2014-2020, xarray Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# xarray bundles portions of pandas, NumPy and Seaborn, all of which are available
# under a "3-clause BSD" license:
# - pandas: setup.py, xarray/util/print_versions.py
# - NumPy: xarray/core/npcompat.py
# - Seaborn: _determine_cmap_params in xarray/core/plot/utils.py
#
# xarray also bundles portions of CPython, which is available under the "Python
# Software Foundation License" in xarray/core/pycompat.py.
#
# xarray uses icons from the icomoon package (free version), which is
# available under the "CC BY 4.0" license.
#
# The full text of these licenses are included in the licenses directory.
class ExplicitlyIndexed:
    """Mixin to mark support for Indexer subclasses in indexing."""

    __slots__ = ()


class ExplicitIndexer:
    """Base class for explicit indexer objects.

    ExplicitIndexer objects wrap a tuple of values given by their ``tuple``
    property. These tuples should always have length equal to the number of
    dimensions on the indexed array.

    Do not instantiate BaseIndexer objects directly: instead, use one of the
    sub-classes BasicIndexer, OuterIndexer or VectorizedIndexer.
    """

    __slots__ = ("_key",)

    def __init__(self, key):
        if type(self) is ExplicitIndexer:
            raise TypeError("cannot instantiate base ExplicitIndexer objects")
        self._key = tuple(key)

    @property
    def tuple(self):
        return self._key

    def __repr__(self):
        return f"{type(self).__name__}({self.tuple})"



# This file is part of IMASPy.
#
# IMASPy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IMASPy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IMASPy.  If not, see <https://www.gnu.org/licenses/>.
import importlib
import logging
import os

from imaspy.backends.common import WritableIMASDataStore
from imaspy.backends.file_manager import DummyFileManager
from imaspy.imas_ual_env_parsing import (
    build_UAL_package_name,
    parse_UAL_version_string,
    sanitise_UAL_symver,
)

root_logger = logging.getLogger("imaspy")
logger = root_logger
logger.setLevel(logging.WARNING)

try:
    # TODO: change to new structure
    from imas.imasdef import *
except ImportError:
    logger.warning(
        "imasdef unavailable, might give trouble when using IMAS-AL related functionality"
    )
else:
    AL_BACKENDS = {
        NO_BACKEND: "NO BACKEND",
        ASCII_BACKEND: "ASCII BACKEND",
        MDSPLUS_BACKEND: "MDSPLUS BACKEND",
        HDF5_BACKEND: "HDF5 BACKEND",
        MEMORY_BACKEND: "MEMORY BACKEND",
        UDA_BACKEND: "UDA BACKEND",
    }
    PULSE_ACTIONS = {
        OPEN_PULSE: "OPEN PULSE",
        FORCE_OPEN_PULSE: "FORCE OPEN PULSE",
        CREATE_PULSE: "CREATE PULSE",
        FORCE_CREATE_PULSE: "FORCE CREATE PULSE",
        CLOSE_PULSE: "CLOSE PULSE",
        ERASE_PULSE: "ERASE PULSE",
    }


def _find_user_name():
    """ Find user name as needed by UAL. """
    if "USER" in os.environ:
        user_name = os.environ["USER"]
    else:
        raise Exception("Could not determine user_name automatically.")
    return user_name


def _find_data_version():
    """ Find IMAS data version. """
    if "IMAS_VERSION" in os.environ:
        data_version = os.environ["IMAS_VERSION"]
    else:
        raise Exception("Could not determine version of data automatically.")
    return data_version


def _find_ual_version():
    """ Find UAL LL data version. """
    if "UAL_VERSION" in os.environ:
        ual_version = os.environ["UAL_VERSION"]
    else:
        raise Exception("Could not determine UAL version automatically.")
    return ual_version


class ALError(Exception):
    ERRORCODES = {
        -2: "Context exception",
        -3: "Backend exception",
        -4: "Abstract low level layer exception",
        -1: "Uknown error",
        None: "errorStatus not passed",
    }

    def __init__(self, message, errorStatus=None):
        self.message = message
        self.errorStatus = errorStatus
        print(self.al_message)

    @property
    def al_message(self):
        return self.ERRORCODES[self.errorStatus]

    def __str__(self):
        exception_message = "{!s}\nAL message='{!s}'".format(
            self.message, self.al_message
        )
        return exception_message


class PulseNotFoundError(ALError):
    pass


def get_user_db_directory(user=None):
    """Get the IMAS database directory root for the user.
    If user is omitted, the current user is used."""
    if not user:
        user = os.getlogin()
    if user == "public":
        publichome = os.getenv("IMAS_HOME")
        if publichome is None:
            raise Exception("Environment variable IMAS_HOME is not defined. Quitting.")
        return publichome + "/shared/imasdb"
    else:
        return os.path.expanduser("~" + user + "/public/imasdb")


class UALFile:
    """ A UAL file-like object. """

    # From the python glossary:

    # An object exposing a file-oriented API (with methods such as read()
    # or write()) to an underlying resource. Depending on the way it was created,
    # a file object can mediate access to a real on-disk file or to another type
    # of storage or communication device (for example standard input/output,
    # in-memory buffers, sockets, pipes, etc.). File objects are also called
    # file-like objects or streams.
    #
    # There are actually three categories of file objects: raw binary files,
    # buffered binary files and text files. Their interfaces are defined in the
    # io module. The canonical way to create a file object is by using the open()
    # function.

    # We implement here the needed methods and properties as needed
    # We use delegation instead of inheriting from io.IOBase
    #
    def __init__(
        self,
        backend_id,
        db_name,
        shot,
        run,
        user_name,
        data_version,
        ual_version_string,
        mode="r",
        options="",
    ):
        # Check sanity of input arguments
        if backend_id == MDSPLUS_BACKEND:
            if shot < 1:
                raise Exception("Shot should be 1 or higher")
            if run > 99999:
                raise Exception("Shot should be 99999 or lower")

        ual_patch_version, steps_from_version, ual_commit = parse_UAL_version_string(
            ual_version_string
        )

        safe_ual_patch_version = sanitise_UAL_symver(ual_patch_version)
        ual_ext_module_name = build_UAL_package_name(safe_ual_patch_version, ual_commit)

        if backend_id not in AL_BACKENDS:
            raise Exception(
                "Given backend_id '{!s}' not in allowed backends".format(backend_id)
            )

        backend_str = AL_BACKENDS[backend_id]

        self.db_name = db_name
        self.shot = shot
        self.run = run
        self.user_name = user_name
        self.data_version = data_version
        self.options = options
        self.ual_module_name = ual_ext_module_name
        self.backend_id = backend_id

        # Import imaspy UAL library
        ull = importlib.import_module(ual_ext_module_name)

        # Begin the pulse action
        status, idx = ull.ual_begin_pulse_action(
            backend_id, shot, run, user_name, db_name, data_version
        )
        pulse_action_state = "{!r},{!r},{!r},{!r},{!r}".format(
            backend_id, shot, run, user_name, db_name, data_version
        )
        if status != 0:
            raise ALError(
                "Error calling ual_begin_pulse_action({!r})".format(pulse_action_state),
                status,
            )

        # OPEN_PULSE: Openes the access to the data only if the Data Entry
        #    exists, returns error otherwise
        # FORCE_OPEN_PULSE: Opens access to the data, creates the Data Entry
        #    if it does not exist yet
        # CREATE_PULSE: Creates a new empty Data Entry (returns error if
        #    Data Entry already exists) and opens it at the same time
        # FORCE_CREATE_PULSE: Creates an empty Data Entry (overwrites if
        #    Data Entry already exists) and opens it at the same time
        if mode == "r":
            status = ull.ual_open_pulse(idx, OPEN_PULSE, options)
            if status != 0:
                raise PulseNotFoundError("No such pulse {!r}".format(self), status)
        elif mode == "w":
            status = ull.ual_open_pulse(idx, FORCE_CREATE_PULSE, options)
        elif mode == "a":
            status = ull.ual_open_pulse(idx, FORCE_OPEN_PULSE, options)
        elif mode == "x":
            status = ull.ual_open_pulse(idx, CREATE_PULSE, options)
        else:
            raise ValueError("Invalid mode: {!r}".format(mode))

        if status != 0:
            raise ALError(
                "Error calling ull.ual_open_pulse({!r},{!r},{!r}).\n"
                "Pulse action state was ({!s}).\n"
                "Backend was {!r}".format(
                    idx, OPEN_PULSE, options, pulse_action_state, backend_str
                ),
                status,
            )

        self.closed = False
        self._context_idx = idx
        self._attrs_locked = True

    @property
    def readable(self):
        # An UALFile is automatically readable if it is open
        return not self.closed

    @property
    def writeable(self):
        # An UALFile is automatically writable if it is open
        return not self.closed

    @classmethod
    def open(
        cls,
        backend_id,
        db_name,
        shot,
        run,
        mode="r",
        user_name=None,
        data_version=None,
        options="",
        ual_version=None,
    ):
        # Try to automatically find not-given kwargs
        user_name = user_name or _find_user_name()
        data_version = data_version or _find_data_version()
        ual_version = ual_version or _find_ual_version()
        return cls(
            backend_id,
            db_name,
            shot,
            run,
            user_name,
            data_version,
            ual_version,
            mode=mode,
            options=options,
        )

    @classmethod
    def create(cls, options=""):
        raise Exception(
            "Create a using {!s}.open(*args, mode='w', **kwargs)".format(cls)
        )

    def __repr__(self):
        return "%s(context=%r)" % (type(self).__name__, self._context_idx)

    def __str__(self):
        return "%s(shot=%r, run=%r)" % (type(self).__name__, self.shot, self.run)

    def close(self, options=None):
        if self.db_ctx != -1:
            old_attrs_locked = self._attrs_locked
            self._attrs_locked = False
            ull.ual_close_pulse(self.db_ctx, CLOSE_PULSE, options)
            self.db_ctx = -1
            self.closed = True
            self._attrs_locked = old_attrs_locked

    def filepath(self):
        path = "/"
        if self.backend_id in (MDSPLUS_BACKEND,):
            # MDSPlus Pulsefules come are named name-of-the-tree_shot-specifier
            # name-of-the-tree is always ids for these file
            # shot-specifier has special values:
            # * -1 - model
            # * 0 - current shot
            # * >1 - pulse files
            # Our UALFile should be a 'pulse file', so >1
            # The last four digits are the run number.
            home = os.path.expanduser("~{!s}".format(self.user_name))
            dbdir = get_user_db_directory(self.user_name)
            mdsplusdir = os.path.join(dbdir, self.db_name, self.data_version)
            treedir = os.path.join(mdsplusdir, str(int(self.run / 10000)))
            run_string = str(self.run % 10000)
            if self.shot == 0:
                path = os.path.join(treedir, "ids_" + run_string.zfill(3) + ".*")
            else:
                path = os.path.join(
                    treedir, "ids_" + str(self.shot) + run_string.zfill(4) + ".*"
                )
        return path

    def __setattr__(self, key, value):
        # Prevent user from trying to change what this file points to by
        # changing its attributes
        if hasattr(self, "_attributes_locked") and self._attributes_locked:
            raise AttributeError(
                "attribute '{!s}' of '{!s}' objects is not writable".format(
                    key, type(self)
                )
            )
        else:
            super().__setattr__(key, value)


# This interface _heavily_ borrows from NetCDF4DataStore, as it is
# very similar to how IDSs in UAL are represented. However, all
# currently not needed functionality has been removed.
class UALDataStore(WritableIMASDataStore):
    """Store for reading and writing data via the UAL Low-Level interface."""

    __slots__ = (
        "autoclose",
        "format",
        "is_remote",
        "lock",
        "_filename",
        "_group",
        "_manager",
        "_mode",
    )

    def __init__(
        self,
        manager,
        # group=None,
        # mode=None,
        # lock=NETCDF4_PYTHON_LOCK, Do use locking
        # autoclose=False
    ):
        self._manager = manager
        # self._group = group
        # self._mode = mode
        # self.format = self.ds.data_model
        self._filename = self.ds.filepath()
        # self.is_remote = is_remote_uri(self._filename)
        self.is_remote = False
        self.lock = False
        self.autoclose = False

    @classmethod
    def open(
        cls,
        backend_id,
        db_name,
        shot,
        run,
        user_name=None,
        data_version=None,
        mode="r",
        # format="NETCDF4", # There is only one format, UAL
        # group=None, # All sub-level access is handeled by the UAL directly
        # clobber=False, # Other level?
        # diskless=False, # Handled by UAL?
        # persist=False, # No idea what this does
        # lock=None, # File locking handeled by the UAL
        # lock_maker=None, # File locking handeled by the UAL
        # autoclose=False, # Nope
        options=None,
        ual_version=None,
    ):
        """ Open the UAL store"""
        ual_file = UALFile.open(
            backend_id,
            db_name,
            shot,
            run,
            mode=mode,
            user_name=user_name,
            data_version=data_version,
            options=options,
            ual_version=ual_version,
        )
        manager = DummyFileManager(ual_file)
        return cls(manager)

    def _acquire(self, needs_lock=True):
        """Acquire a reference to this specific UALDataStore"""
        # Get the context from the manager and use it as root
        with self._manager.acquire_context() as root:
            # As we only _allow_ to grab internal structures like this, the
            # following code is easier than in the xarray netCDF4 case
            ds = root
        return ds

    @property
    def ds(self):
        return self._acquire()

    def sync(self):
        self.ds.sync()

    def close(self, **kwargs):
        self._manager.close(**kwargs)

    @property
    def _idx(self):
        pulse_file = self._manager.acquire()
        return pulse_file._context_idx



# This file is part of IMASPy.
#
# IMASPy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IMASPy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IMASPy.  If not, see <https://www.gnu.org/licenses/>.
""" Core IDS classes

Provides the core classes:

* :py:class:`IDSPrimitive`
* :py:class:`IDSRoot`
* :py:class:`IDSStructure`
* :py:class:`IDSStructArray`
* :py:class:`IDSToplevel`
"""

# Set up logging immediately
import copy
import functools
import importlib
import logging
import numbers
import os
import xml.etree.ElementTree as ET

import numpy as np

root_logger = logging.getLogger("imaspy")
logger = root_logger
logger.setLevel(logging.WARNING)


try:
    import imas.hli_utils as hli_utils
    from imas.imasdef import (
        MDSPLUS_BACKEND,
        OPEN_PULSE,
        READ_OP,
        EMPTY_INT,
        FORCE_CREATE_PULSE,
        IDS_TIME_MODE_UNKNOWN,
        IDS_TIME_MODES,
        IDS_TIME_MODE_HOMOGENEOUS,
        IDS_TIME_MODE_HETEROGENEOUS,
        WRITE_OP,
        CHAR_DATA,
        INTEGER_DATA,
        EMPTY_FLOAT,
        DOUBLE_DATA,
        NODE_TYPE_STRUCTURE,
        CLOSE_PULSE,
    )
except ImportError:
    logger.critical("IMASPy _libs could not be imported. UAL not available!")
else:
    # Translation dictionary to go from an ids (primitive) type (without the dimensionality) to a default value
    ids_type_to_default = {
        "STR": "",
        "INT": EMPTY_INT,
        "FLT": EMPTY_FLOAT,
        "FLT_0D": EMPTY_FLOAT,
    }


class ContextStore(dict):
    """Stores global UAL context

    A context is a sort of pointer but to where depends on the type of context:
      - PulseContext: identifies a specific entry in database
      - OperationContext: identifies a specific I/O operation (read/write,
        global/slice, which IDS, etc...) being performed on a specific
        PulseContext
      - ArraystructContext: identifies a array of structure node within the IDS
        for a specific operation

    The rest of the absolute path (from last context to leaf/data) is not stored
    in context but passed directly to ual_read and ual_write LL functions.
    Contexts have a fullPath() method that will return string with pseudo
    fullpath up to this context
    """

    def __setitem__(self, key, value):
        """Store context id (key) and full path (value)

        As context is stored globally within the LL-UAL beyond our reach,
        do not allow for duplicated contexts to be opened.
        """
        if key in self:
            raise Exception(
                "Trying to set context {!s} to {!s}, but was not released. Currently is {!s}".format(
                    key, value, self[key]
                )
            )
        else:
            super().__setitem__(key, value)

    def update(self, ctx, newCtx):
        if ctx not in self:
            raise Exception("Trying to update non-existing context {!s}".format(ctx))
        super().__setitem__(ctx, newCtx)

    def decodeContextInfo(self, ctxLst=None):
        """Decode ual context info to Python-friendly format"""
        if ctxLst is None:
            ctxLst = self.keys()
        elif ctxLst is not None and not isinstance(cnxLst, list):
            ctxLst = list(ctxLst)
        if any(not np.issubdtype(type(ctx), np.integer) for ctx in ctxLst):
            raise ValuError("ctx identifier should be an integer")
        for ctx in ctxLst:
            # This seems to cause memory corruption
            # Sometimes..
            contextInfo = self._ull.ual_context_info(ctx)
            infoCopy = (contextInfo + ".")[:-1]
            info = {}
            for line in infoCopy.split("\n"):
                if line == "":
                    continue
                key, val = line.split("=")
                info[key.strip()] = val.strip()
            print("ctx", ctx, info)


# Keep the context store on the module level
# TODO: Decide if this is the place to put it. How 'global' is 'global'?
context_store = ContextStore()


class ALException(Exception):
    def __init__(self, message, errorStatus=None):
        if errorStatus is not None:
            Exception.__init__(self, message + "\nError status=" + str(errorStatus))
        else:
            Exception.__init__(self, message)


def loglevel(func):
    """Generate a decorator for setting the logger level on a function"""

    @functools.wraps(func)
    def loglevel_decorator(*args, **kwargs):
        verbosity = kwargs.pop("verbosity", None)
        if verbosity is not None:
            old_log_level = logger.level
            if verbosity >= 1:
                logger.setLevel(logging.INFO)
            if verbosity >= 2:
                logger.setLevel(logging.DEBUG)
        value = func(*args, **kwargs)
        if verbosity is not None:
            logger.setLevel(old_log_level)
        return value

    return loglevel_decorator


class IDSMixin:
    @loglevel
    def getRelCTXPath(self, ctx):
        """ Get the path relative to given context from an absolute path"""
        # This could be replaced with the fullPath() method provided by the LL-UAL
        if self.path.startswith(context_store[ctx]):
            # If the given path indeed starts with the context path. This should
            # always be the case. Grab the part of the path _after_ the context
            # path string
            if context_store[ctx] == "/":
                # The root context is special, it does not have a slash before
                rel_path = self.path[len(context_store[ctx]) :]
            else:
                rel_path = self.path[len(context_store[ctx]) + 1 :]
            split = rel_path.split("/")
            try:
                # Check if the first part of the path is a number. If it is,
                # strip it, it is implied by context
                int(split[0])
            except (ValueError):
                pass
            else:
                # Starts with numeric, strip. Is captured in context
                # TODO: Might need to be recursive.
                rel_path = "/".join(split[1:])
        else:
            raise Exception("Could not strip context from absolute path")
        # logger.debug('Got context {!s} with abspath {!s}, relpath is {!s}'.format(ctx, self.path, rel_path))
        return rel_path

    def getTimeBasePath(self, homogeneousTime, ignore_nbc_change=1):
        strTimeBasePath = ""
        # Grab timebasepath from the coordinates.
        # TODO: In some cases the timebasepath is stored in the XML directly.
        #       What has priority in case it conflicts? Regardless, this is not
        #       handled by imaspy atm
        if self._coordinates != {}:
            if (
                self._coordinates["coordinate1"].endswith("time")
                and "coordinate2" not in self._coordinates
            ):
                # Should Walk up the tree
                # Just stupid copy for now
                # strTimeBasePath = self._coordinates['coordinate1']
                if homogeneousTime == IDS_TIME_MODE_HOMOGENEOUS:
                    strTimeBasePath = "/time"
                elif homogeneousTime == IDS_TIME_MODE_HETEROGENEOUS:
                    strTimeBasePath = self.getAOSPath(ignore_nbc_change) + "/time"
                else:
                    raise ALException(
                        "Unexpected call to function getTimeBasePath(cls, homogeneousTime) with undefined homogeneous time."
                    )
                pass
            elif (
                self._coordinates["coordinate1"] == "1...N"
                and "coordinate2" not in self._coordinates
            ):
                # If variable only depends on 1...N, no timebasepath
                pass
            else:
                # Stub for explicit handling of other cases
                pass

        return strTimeBasePath

    def getAOSPath(self, ignore_nbc_change=1):
        # TODO: Fix in case it gives trouble
        # This is probably wrong! Should walk up the tree
        return self._name

    @property
    def path(self):
        """Build absolute path from node to root"""
        # Probably superseded by the property below
        my_path = self._name
        if hasattr(self, "_parent"):
            my_path = self._parent.path + "/" + my_path
        return my_path

    @property
    def path(self):
        """Build absolute path from node to root"""
        my_path = self._name
        if hasattr(self, "_parent"):
            if isinstance(self._parent, IDSStructArray):
                my_path = "{!s}/{!s}".format(
                    self._parent.path, self._parent.value.index(self)
                )
            else:
                my_path = self._parent.path + "/" + my_path
        return my_path

    @property
    def _ull(self):
        if not hasattr(self, "_parent"):
            raise Exception("ULL directly connected to {!s}".format(self))
        return self._parent._ull


class IDSPrimitive(IDSMixin):
    """IDS leaf node

    Represents actual data. Examples are (arrays of) strings, floats, integers.
    Lives entirely in-memory until 'put' into a database.
    """

    @loglevel
    def __init__(
        self, name, ids_type, ndims, parent=None, value=None, coordinates=None
    ):
        """Initialize IDSPrimitive

        args:
          - name: Name of the leaf node. Will be used in path generation when
                  stored in DB
          - ids_type: String representing the IDS type. Will be used to convert
                      to Python equivalent
          - ndims: Dimensionality of data

        kwargs:
          - parent: Parent node of this leaf. Can be anything with a _path attribute.
                    Will be used in path generation when stored in DB
          - value: Value to fill the leaf with. Can be anything castable by
                   IDSPrimitive.cast_value. If not given, will be filled by
                   default data matching given ids_type and ndims
          - coordinates: Data coordinates of the node
        """
        if ids_type != "STR" and ndims != 0 and self.__class__ == IDSPrimitive:
            raise Exception(
                "{!s} should be 0D! Got ndims={:d}. Instantiate using IDSNumericArray instead".format(
                    self.__class__, ndims
                )
            )
        if ndims == 0:
            self._default = ids_type_to_default[ids_type]
        else:
            self._default = np.full((1,) * ndims, ids_type_to_default[ids_type])
        if value is None:
            value = self._default
        self._ids_type = ids_type
        self._ndims = ndims
        self._name = name
        self._parent = parent
        self._coordinates = coordinates
        self.value = value

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, setter_value):
        if isinstance(setter_value, type(self)):
            # No need to cast, just overwrite contained value
            if (
                setter_value._ids_type == self._ids_type
                and setter_value._ndims == self._ndims
            ):
                self.__value = setter_value.value
            # Can we cast the internal value to a valid value?
            else:
                self.__value = self.cast_value(setter_value.value)
        else:
            self.__value = self.cast_value(setter_value)

    def cast_value(self, value):
        # Cast list-likes to arrays
        if isinstance(value, (list, tuple)):
            value = np.array(value)

        # Cast values to their IDS-python types
        if self._ids_type == "STR" and self._ndims == 0:
            value = str(value)
        elif self._ids_type == "INT" and self._ndims == 0:
            value = int(value)
        elif self._ids_type == "FLT" and self._ndims == 0:
            value = float(value)
        elif self._ndims >= 1:
            if self._ids_type == "FLT":
                value = np.array(value, dtype=np.float64)
            elif self._ids_type == "INT":
                value = np.array(value, dtype=np.int64)
            else:
                logger.critical(
                    "Unknown numpy type {!s}, cannot convert from python to IDS type".format(
                        value.dtype
                    )
                )
                raise Exception
        else:
            logger.critical(
                "Unknown python type {!s}, cannot convert from python to IDS type".format(
                    type(value)
                )
            )
            raise Exception
        return value

    @loglevel
    def put(self, ctx, homogeneousTime):
        """Put data into UAL backend storage format

        Does minor sanity checking before calling the cython backend.
        Tries to dynamically build all needed information for the UAL.
        """
        if self._name is None:
            raise Exception("Location in tree undefined, cannot put in database")
        # Convert imaspy ids_type to ual scalar_type
        if self._ids_type == "INT":
            scalar_type = 1
        elif self._ids_type == "FLT":
            scalar_type = 2
        elif self._ids_type == "CPX":
            scalar_type = 3

        # Check sanity of given data
        if self._ids_type in ["INT", "FLT", "CPX"] and self._ndims == 0:
            data = hli_utils.HLIUtils.isScalarFinite(self.value, scalar_type)
        elif self._ids_type in ["INT", "FLT", "CPX"]:
            if not hli_utils.HLIUtils.isTypeValid(self.value, self._name, "NP_ARRAY"):
                raise Exception
            data = hli_utils.HLIUtils.isFinite(self.value, scalar_type)
        else:
            data = self.value

        # Do not write if data is the same as the default of the leaf node
        if np.all(data == self._default):
            return

        dbg_str = " " + " " * self.depth + "- " + self._name
        dbg_str += (" {:" + str(max(0, 53 - len(dbg_str))) + "s}").format(
            "(" + str(data) + ")"
        )
        # Call signature
        # ual_write_data(ctx, pyFieldPath, pyTimebasePath, inputData, dataType=0, dim = 0, sizeArray = np.empty([0], dtype=np.int32))
        data_type = self._ull._getDataType(data)

        # Strip context from absolute path
        rel_path = self.getRelCTXPath(ctx)
        # TODO: Check ignore_nbc_change
        strTimeBasePath = self.getTimeBasePath(homogeneousTime)

        logger.info("{:54.54s} write".format(dbg_str))
        logger.debug(
            "   {:50.50s} write".format("/".join([context_store[ctx], rel_path]))
        )
        status = self._ull.ual_write_data(
            ctx, rel_path, strTimeBasePath, data, dataType=data_type, dim=self._ndims
        )
        if status != 0:
            raise ALException('Error writing field "{!s}"'.format(self._name))

    @loglevel
    def get(self, ctx, homogeneousTime):
        """Get data from UAL backend storage format

        Tries to dynamically build all needed information for the UAL.
        Does currently _not_ set value of the leaf node, this is handled
        by the IDSStructure.
        """
        # Strip context from absolute path
        strNodePath = self.getRelCTXPath(ctx)
        strTimeBasePath = self.getTimeBasePath(homogeneousTime)
        if self._ids_type == "STR" and self._ndims == 0:
            status, data = self._ull.ual_read_data_string(
                ctx, strNodePath, strTimeBasePath, CHAR_DATA, 1
            )
        elif self._ids_type == "INT" and self._ndims == 0:
            status, data = self._ull.ual_read_data_scalar(
                ctx, strNodePath, strTimeBasePath, INTEGER_DATA
            )
        elif self._ids_type == "FLT" and self._ndims == 0:
            status, data = self._ull.ual_read_data_scalar(
                ctx, strNodePath, strTimeBasePath, DOUBLE_DATA
            )
        elif self._ids_type == "FLT" and self._ndims > 0:
            status, data = self._ull.ual_read_data_array(
                ctx, strNodePath, strTimeBasePath, DOUBLE_DATA, self._ndims
            )
        elif self._ids_type == "INT" and self._ndims > 0:
            status, data = self._ull.ual_read_data_array(
                ctx, strNodePath, strTimeBasePath, INTEGER_DATA, self._ndims
            )
        else:
            logger.critical(
                "Unknown type {!s} ndims {!s} of field {!s}, skipping for now".format(
                    self._ids_type, self._ndims, self._name
                )
            )
            status = data = None
        return status, data

    @property
    def depth(self):
        """Calculate the depth of the leaf node"""
        my_depth = 0
        if hasattr(self, "_parent"):
            my_depth += self._parent.depth
        return my_depth

    def __repr__(self):
        return '%s("%s", %r)' % (type(self).__name__, self._name, self.value)

    @property
    def data_type(self):
        """Combine imaspy ids_type and ndims to UAL data_type"""
        return "{!s}_{!s}D".format(self._ids_type, self._ndims)


def create_leaf_container(name, data_type, **kwargs):
    """Wrapper to create IDSPrimitive/IDSNumericArray from IDS syntax"""
    if data_type == "int_type":
        ids_type = "INT"
        ndims = 0
    elif data_type == "flt_type":
        ids_type = "FLT"
        ndims = 0
    elif data_type == "flt_1d_type":
        ids_type = "FLT"
        ndims = 1
    else:
        ids_type, ids_dims = data_type.split("_")
        ndims = int(ids_dims[:-1])
    if ndims == 0:
        leaf = IDSPrimitive(name, ids_type, ndims, **kwargs)
    else:
        if ids_type == "STR":
            # Array of strings should behave more like lists
            # this is an assumption on user expectation!
            leaf = IDSPrimitive(name, ids_type, ndims, **kwargs)
        else:
            leaf = IDSNumericArray(name, ids_type, ndims, **kwargs)
    return leaf


class IDSNumericArray(IDSPrimitive, np.lib.mixins.NDArrayOperatorsMixin):
    def __str__(self):
        return self.value.__str__()

    # One might also consider adding the built-in list type to this
    # list, to support operations like np.add(array_like, list)
    _HANDLED_TYPES = (np.ndarray, numbers.Number)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        out = kwargs.get("out", ())
        for x in inputs + out:
            # Only support operations with instances of _HANDLED_TYPES.
            # Use ArrayLike instead of type(self) for isinstance to
            # allow subclasses that don't override __array_ufunc__ to
            # handle ArrayLike objects.
            if not isinstance(x, self._HANDLED_TYPES + (IDSPrimitive,)):
                return NotImplemented

        # Defer to the implementation of the ufunc on unwrapped values.
        inputs = tuple(x.value if isinstance(x, IDSPrimitive) else x for x in inputs)
        if out:
            kwargs["out"] = tuple(
                x.value if isinstance(x, IDSPrimitive) else x for x in out
            )
        result = getattr(ufunc, method)(*inputs, **kwargs)

        if type(result) is tuple:
            # multiple return values
            return tuple(
                type(self)(self._name, self._ids_type, self._ndims, value=x)
                for x in result
            )
        elif method == "at":
            # no return value
            return None
        else:
            # one return value
            return type(self)(self._name, self._ids_type, self._ndims, value=result)


class IDSRoot:
    """ Root of IDS tree. Contains all top-level IDSs """

    depth = 0
    path = ""

    @loglevel
    def __init__(self, s=-1, r=-1, rs=None, rr=None, xml_path=None):
        """Initialize a imaspy IDS tree

        Dynamically build the imaspy IDS tree from the given xml path.
        This does not need necessarily need any associated backend,
        but the structure matches MDSPlus pulsefile. E.g. each Root
        is identified by its shot and run, combining into a UID that
        should be unique per database.
        """
        setattr(self, "shot", s)
        self.shot = s
        self.run = r

        if rs is not None:
            raise NotImplementedError("Setting of reference shot")
        if rr is not None:
            raise NotImplementedError("Setting of reference run")

        # The following attributes relate to the UAL-LL
        self.treeName = "ids"
        self.connected = False
        self.expIdx = -1

        # Parse given xml_path and build imaspy IDS structures
        XMLtreeIDSDef = ET.parse(xml_path)
        root = XMLtreeIDSDef.getroot()
        self._children = []
        logger.info(
            "Generating IDS structures from XML file {!s}".format(
                os.path.abspath(xml_path)
            )
        )
        for ids in root:
            my_name = ids.get("name")
            if my_name is None:
                continue
            if my_name not in ["equilibrium", "nbi"]:
                continue
            logger.debug("{:42.42s} initialization".format(my_name))
            self._children.append(my_name)
            setattr(self, my_name, IDSToplevel(self, my_name, ids))

    # self.equilibrium = IDSToplevel('equilibrium')

    # Do not use this now
    # self.ddunits = DataDictionaryUnits()
    # self.hli_utils = HLIUtils()
    # self.amns_data = amns_data.amns_data()
    # self.barometry = barometry.barometry()
    # etc. etc over all lower level IDSs

    def __str__(self, depth=0):
        space = ""
        for i in range(depth):
            space = space + "\t"

        ret = space + "class ids\n"
        ret = (
            ret
            + space
            + "Shot=%d, Run=%d, RefShot%d RefRun=%d\n"
            % (self.shot, self.run, self.refShot, self.refRun)
        )
        ret = (
            ret
            + space
            + "treeName=%s, connected=%d, expIdx=%d\n"
            % (self.treeName, self.connected, self.expIdx)
        )
        ret = ret + space + "Attribute amns_data\n" + self.amns_data.__str__(depth + 1)
        ret = ret + space + "Attribute barometry\n" + self.barometry.__str__(depth + 1)
        # etc. etc over all lower level IDSs
        return ret

    def __del__(self):
        return 1

    def setShot(self, inShot):
        self.shot = inShot

    def setRun(self, inRun):
        self.run = inRun

    def setRefShot(self, inRefShot):
        self.refShot = inRefShot

    def setRefNum(self, inRefRun):
        self.refRun = inRefRun

    def setTreeName(self, inTreeName):
        self.treeName = inTreeName

    def getShot(self):
        return self.shot

    def getRun(self):
        return self.run

    def getRefShot(self):
        return self.refShot

    def getRefRun(self):
        return self.refRun

    def getTreeName(self):
        return self.treeName

    def isConnected(self):
        return self.connected

    def get_units(self, ids, field):
        return self.ddunits.get_units(ids, field)

    def get_units_parser(self):
        return self.ddunits

    def open_ual_store(
        self,
        user,
        tokamak,
        version,
        backend_type,
        mode="r",
        silent=False,
        options="",
        ual_version=None,
    ):
        from imaspy.backends.ual import UALDataStore

        if silent:
            options += "-silent"

        store = UALDataStore.open(
            backend_type,
            tokamak,
            self.shot,
            self.run,
            user_name=user,
            data_version=version,
            mode=mode,
            options=options,
            ual_version=ual_version,
        )

        # Safe the store internally for magic path detection
        self._data_store = store

        # Do we need to set context like dis?
        self.setPulseCtx(store._idx)
        context_store[store._idx] = "/"

        status = 0
        return store

    def create_env(
        self, user, tokamak, version, silent=False, options="", ual_version=None
    ):
        """Creates a new pulse.

        Parameters
        ----------
        user : string
            Owner of the targeted pulse.
        tokamak : string
            Tokamak name for the targeted pulse.
        version : string
            Data-dictionary major version number for the targeted pulse.
        silent : bool, optional
            Request the lowlevel to be silent (does not print error messages).
        options : string, optional
            Pass additional options to lowlevel.
        ual_version: string, optional
            Specify the UAL version to be used. Use format x.x.x
        """
        store = self.open_ual_store(
            user,
            tokamak,
            version,
            MDSPLUS_BACKEND,
            mode="w",  # This is different per env call
            silent=silent,
            options=options,
            ual_version=ual_version,
        )
        return (0, store._idx)

    def create_env_backend(
        self,
        user,
        tokamak,
        version,
        backend_type,
        silent=False,
        options="",
        ual_version=None,
    ):
        """Creates a new pulse for a UAL supported backend

        Parameters
        ----------
        user : string
            Owner of the targeted pulse.
        tokamak : string
            Tokamak name for the targeted pulse.
        version : string
            Data-dictionary major version number for the targeted pulse.
        backend_type: integer
            One of the backend types (e.g.: MDSPLUS_BACKEND, MEMORY_BACKEND).
        silent : bool, optional
            Request the lowlevel to be silent (does not print error messages).
        options : string, optional
            Pass additional options to lowlevel.
        ual_version: string, optional
            Specify the UAL version to be used. Use format x.x.x
        """
        store = self.open_ual_store(
            user,
            tokamak,
            version,
            backend_type,
            mode="w",  # This is different per env call
            silent=silent,
            options=options,
            ual_version=ual_version,
        )
        return (0, store._idx)

    def open_env(
        self, user, tokamak, version, silent=False, options="", ual_version=None
    ):
        """Opens an existing pulse.

        Parameters
        ----------
        user : string
            Owner of the targeted pulse.
        tokamak : string
            Tokamak name for the targeted pulse.
        version : string
            Data-dictionary major version number for the targeted pulse.
        silent : bool, optional
            Request the lowlevel to be silent (does not print error messages).
        options : string, optional
            Pass additional options to lowlevel.
        ual_version: string, optional
            Specify the UAL version to be used. Use format x.x.x
        """
        store = self.open_ual_store(
            user,
            tokamak,
            version,
            backend_type,
            mode="r",  # This is different per env call
            silent=silent,
            options=options,
            ual_version=ual_version,
        )
        return (0, store._idx)

    def open_env_backend(
        self,
        user,
        tokamak,
        version,
        backend_type,
        silent=False,
        options="",
        ual_version=None,
    ):
        """Opens an existing pulse for UAL supported backend.

        Parameters
        ----------
        user : string
            Owner of the targeted pulse.
        tokamak : string
            Tokamak name for the targeted pulse.
        version : string
            Data-dictionary major version number for the targeted pulse.
        backend_type: integer
            One of the backend types (e.g.: MDSPLUS_BACKEND, MEMORY_BACKEND).
        silent : bool, optional
            Request the lowlevel to be silent (does not print error messages).
        options : string, optional
            Pass additional options to lowlevel.
        ual_version: string, optional
            Specify the UAL version to be used. Use format x.x.x
        """
        store = self.open_ual_store(
            user,
            tokamak,
            version,
            backend_type,
            mode="r",  # This is different per env call
            silent=silent,
            options=options,
            ual_version=ual_version,
        )
        return (0, store._idx)

    def open_public(self, expName, silent=False):
        """Opens a public pulse with the UAL UAD backend. """
        status, idx = self._ull.ual_begin_pulse_action(
            UDA_BACKEND, self.shot, self.run, "", expName, os.environ["IMAS_VERSION"]
        )
        if status != 0:
            return (status, idx)
        opt = ""
        if silent:
            opt = "-silent"
        status = self._ull.ual_open_pulse(idx, OPEN_PULSE, opt)
        if status != 0:
            return (status, idx)
        self.setPulseCtx(idx)
        context_store[idx] = "/"
        return (status, idx)

    def getPulseCtx(self):
        return self.expIdx

    def setPulseCtx(self, ctx):
        # This sets the contexts of the Root. More-or-less a pointer to a specific pulsefile
        self.expIdx = ctx
        self.connected = True
        # Different than before, IDS TopLevels should get the context from their parent directly
        # self.equilibrium.setPulseCtx(ctx)

    def close(self):
        if self.expIdx != -1:
            status = self._ull.ual_close_pulse(self.expIdx, CLOSE_PULSE, "")
            if status != 0:
                return status
            self.connected = False
            self.expIdx = -1
            return status

    def enableMemCache(self):
        return 1

    def disableMemCache(self):
        return 1

    def discardMemCache(self):
        return 1

    def flushMemCache(self):
        return 1

    def getTimes(self, path):
        homogenousTime = IDS_TIME_MODE_UNKNOWN
        if self.expIdx < 0:
            raise ALException("ERROR: backend not opened.")

        # Create READ context
        status, ctx = self._ull.ual_begin_global_action(self.expIdx, path, READ_OP)
        if status != 0:
            raise ALException("Error calling ual_begin_global_action() for ", status)

        # Check homogeneous_time
        status, homogenousTime = self._ull.ual_read_data(
            ctx, "ids_properties/homogeneous_time", "", INTEGER_DATA, 0
        )
        if status != 0:
            raise ALException("ERROR: homogeneous_time cannot be read.", status)

        if homogenousTime == IDS_TIME_MODE_UNKNOWN:
            status = self._ull.ual_end_action(ctx)
            if status != 0:
                raise ALException("Error calling ual_end_action().", status)
            return 0, []
        # Heterogeneous IDS #
        if homogenousTime == IDS_TIME_MODE_HETEROGENEOUS:
            status = self._ull.ual_end_action(ctx)
            if status != 0:
                raise ALException("ERROR calling ual_end_action().", status)
            return 0, [np.NaN]

        # Time independent IDS #
        if homogenousTime == IDS_TIME_MODE_INDEPENDENT:
            status = self._ull.ual_end_action(ctx)
            if status != 0:
                raise ALException("ERROR calling ual_end_action().", status)
            return 0, [np.NINF]

        # Get global time
        timeList = []
        status, data = self._ull.ual_read_data_array(
            ctx, "time", "/time", DOUBLE_DATA, 1
        )
        if status != 0:
            raise ALException("ERROR: Time vector cannot be read.", status)
        if data is not None:
            timeList = data
        status = self._ull.ual_end_action(ctx)
        if status != 0:
            raise ALException("ERROR calling ual_end_action().", status)
        return status, timeList

    @property
    def _ull(self):
        ctx_path = context_store[self.expIdx]
        if ctx_path != "/":
            raise Exception("{!s} context does not seem to be toplevel".format(self))
        ual_file = self._data_store._manager.acquire()
        ull = importlib.import_module(ual_file.ual_module_name)
        return ull


class IDSStructure(IDSMixin):
    """IDS structure node

    Represents a node in the IDS tree. Does not itself contain data,
    but contains references to leaf nodes with data (IDSPrimitive) or
    other node-like structures, for example other IDSStructures or
    IDSStructArrays
    """

    _MAX_OCCURRENCES = None

    def getNodeType(self):
        raise NotImplementedError("{!s}.getNodeType()".format(self))
        return NODE_TYPE_STRUCTURE

    # def __deepcopy__(self, memo):
    #    raise NotImplementedError

    # def __copy__(self):
    #    raise NotImplementedError
    @loglevel
    def __init__(self, parent, name, structure_xml):
        """Initialize IDSStructure from XML specification

        Initializes in-memory an IDSStructure. The XML should contain
        all direct descendants of the node. To avoid duplication,
        none of the XML structure is saved directly, so this transformation
        might be irreversible.

        Args:
          - parent: Parent structure. Can be anything, but at database write
                    time should be something with a path attribute
          - name: Name of the node itself. Will be used in path generation when
                  stored in DB
          - structure_xml: Object describing the structure of the IDS. Usually
                           an instance of `xml.etree.ElementTree.Element`
        """
        # To ease setting values at this stage, do not try to cast values
        # to canonical forms
        self._convert_ids_types = False
        self._name = name
        self._base_path = name
        self._children = []  # Store the children as a list of strings.
        # As we cannot restore the parent from just a string, save a reference
        # to the parent. Take care when (deep)copying this!
        self._parent = parent
        self._coordinates = {
            attr: structure_xml.attrib[attr]
            for attr in structure_xml.attrib
            if attr.startswith("coordinate")
        }
        # Loop over the direct descendants of the current node.
        # Do not loop over grandchildren, that is handled by recursiveness.
        for child in structure_xml.getchildren():
            my_name = child.get("name")
            dbg_str = " " * (self.depth + 1) + "- " + my_name
            logger.debug("{:42.42s} initialization".format(dbg_str))
            self._children.append(my_name)
            # Decide what to do based on the data_type attribute
            my_data_type = child.get("data_type")
            if my_data_type == "structure":
                child_hli = IDSStructure(self, my_name, child)
                setattr(self, my_name, child_hli)
            elif my_data_type == "struct_array":
                child_hli = IDSStructArray(self, my_name, child)
                setattr(self, my_name, child_hli)
            else:
                # If it is not a structure or struct_array, it is probably a
                # leaf node. Just naively try to generate one
                tbp = child.get("timebasepath")
                if tbp is not None:
                    logger.critical(
                        "Found a timebasepath of {!s}! Should not happen".format(tbp)
                    )
                coordinates = {
                    attr: child.attrib[attr]
                    for attr in child.attrib
                    if attr.startswith("coordinate")
                }
                setattr(
                    self,
                    my_name,
                    create_leaf_container(
                        my_name, my_data_type, parent=self, coordinates=coordinates
                    ),
                )
        # After initialization, always try to convert setting attributes on this structure
        self._convert_ids_types = True

    @property
    def depth(self):
        """Calculate the depth of the leaf node"""
        my_depth = 0
        if hasattr(self, "_parent"):
            my_depth += 1 + self._parent.depth
        return my_depth

    def copyValues(self, ids):
        """ Not sure what this should do. Well, copy values of a structure!"""
        raise NotImplementedError("{!s}.copyValues(ids)".format(self))

    def __str__(self):
        return '%s("%s")' % (type(self).__name__, self._name)

    def __getitem__(self, key):
        keyname = str(key)
        return getattr(self, keyname)

    def __setitem__(self, key, value):
        keyname = str(key)
        self.__setattr__(keyname, value)

    def __setattr__(self, key, value):
        """
        'Smart' setting of attributes. To be able to warn the user on imaspy
        IDS interaction time, instead of on database put time
        Only try to cast user-facing attributes, as core developers might
        want to always bypass this mechanism (I know I do!)
        """
        # TODO: Check if this heuristic is sufficient
        if (
            not key.startswith("_")
            and hasattr(self, "_convert_ids_types")
            and self._convert_ids_types
        ):
            # Convert IDS type on set time. Never try this for hidden attributes!
            if hasattr(self, key):
                attr = getattr(self, key)
            else:
                # Structure does not exist. It should have been pre-generated
                raise NotImplementedError("generating new structure from scratch")
                attr = create_leaf_container(key, no_data_type_I_guess, parent=self)
            if isinstance(attr, IDSStructure) and not isinstance(value, IDSStructure):
                raise Exception(
                    "Trying to set structure field {!s} with non-structure.".format(key)
                )

            try:
                attr.value = value
            except Exception as ee:
                raise
            else:
                object.__setattr__(self, key, attr)
        else:
            object.__setattr__(self, key, value)

    @loglevel
    def readTime(self, occurrence):
        raise NotImplementedError("{!s}.readTime(occurrence)".format(self))
        time = []
        path = None
        if occurrence == 0:
            path = self._name
        else:
            path = self._name + "/" + str(occurrence)

        status, ctx = self._ull.ual_begin_global_action(self._idx, path, READ_OP)
        if status != 0:
            raise ALException(
                "Error calling ual_begin_global_action() in readTime() operation",
                status,
            )

        status, time = self._ull.ual_read_data_array(
            ctx, "time", "/time", DOUBLE_DATA, 1
        )
        if status != 0:
            raise ALException("ERROR: TIME cannot be read.", status)
        status = self._ull.ual_end_action(ctx)
        if status != 0:
            raise ALException(
                "Error calling ual_end_action() in readTime() operation", status
            )
        return time

    @loglevel
    def get(self, ctx, homogeneousTime):
        """Get data from UAL backend storage format and overwrite data in node

        Tries to dynamically build all needed information for the UAL.
        """
        if len(self._children) == 0:
            logger.warning(
                'Trying to get structure "{!s}" with 0 children'.format(self._name)
            )
        for child_name in self._children:
            dbg_str = " " * self.depth + "- " + child_name
            logger.debug("{:53.53s} get".format(dbg_str))
            child = getattr(self, child_name)
            if isinstance(child, IDSStructure):
                child.get(ctx, homogeneousTime)
                continue  # Nested struct will handle setting attributes
            if isinstance(child, IDSPrimitive):
                status, data = child.get(ctx, homogeneousTime)
            else:
                logger.critical(
                    "Unknown type {!s} for field {!s}! Skipping".format(
                        type(child), child_name
                    )
                )
            if status == 0 and data is not None:
                setattr(self, child_name, data)
            elif status != 0:
                logger.critical(
                    "Unable to get simple field {!s}, UAL return code {!s}".format(
                        child_name, status
                    )
                )
            else:
                logger.debug(
                    "Unable to get simple field {!s}, seems empty".format(child_name)
                )

    @loglevel
    def getSlice(
        self, time_requested, interpolation_method, occurrence=0, data_store=None
    ):
        # Retrieve full IDS data from the open database.
        raise NotImplementedError(
            "{!s}.getSlice(time_requested, interpolation_method, occurrence=0, data_store=None)".format(
                self
            )
        )

    @loglevel
    def _getData(self, ctx, homogeneousTime, nodePath, analyzeTime):
        """ A deeped way of getting data?? using 'traverser' whatever that is """
        raise NotImplementedError(
            "{!s}._getData(ctx, homogeneousTime, nodePath, analyzeTime)".format(self)
        )

    @loglevel
    def put(self, ctx, homogeneousTime):
        """Put data into UAL backend storage format

        As all children _should_ support being put, just call `put` blindly.
        """
        if len(self._children) == 0:
            logger.warning(
                "Trying to put structure {!s} without children to data store".format(
                    self._name
                )
            )
        for child_name in self._children:
            child = getattr(self, child_name)
            dbg_str = " " * self.depth + "- " + child_name
            if child is not None:
                if not isinstance(child, IDSPrimitive):
                    logger.debug("{:53.53s} put".format(dbg_str))
                child.put(ctx, homogeneousTime)

    @loglevel
    def putSlice(self, ctx, homogeneousTime):
        # Store IDS data time slice to the open database.
        raise NotImplementedError("{!s}.putSlice(ctx, homogeneousTime)".format(self))

    def setPulseCtx(self, ctx):
        raise DeprecationWarning(
            "IDSs should not set context directly, set on Root node instead"
        )

    def getPulseCtx(self):
        raise DeprecationWarning(
            "IDSs should not set context directly, set on Root node instead"
        )

    @loglevel
    def delete(self, ctx):
        """Delete data from UAL backend storage"""
        for child_name in self._children:
            child = getattr(self, child_name)
            dbg_str = " " * self.depth + "- " + child_name
            logger.debug("{:53.53s} del".format(dbg_str))
            rel_path = child.getRelCTXPath(ctx)
            if isinstance(child, (IDSStructArray, IDSPrimitive)):
                status = self._ull.ual_delete_data(ctx, rel_path)
                if status != 0:
                    raise ALException(
                        'ERROR: ual_delete_data failed for "{!s}". Status code {!s}'.format(
                            rel_path + "/" + child_name
                        ),
                        status,
                    )
            else:
                status = child.delete(ctx)
                if status != 0:
                    raise ALException(
                        'ERROR: delete failed for "{!s}". Status code {!s}'.format(
                            rel_path + "/" + child_name
                        ),
                        status,
                    )
        return 0


class IDSStructArray(IDSStructure, IDSMixin):
    """IDS array of structures (AoS) node

    Represents a node in the IDS tree. Does not itself contain data,
    but contains references to IDSStructures
    """

    def getAOSPath(self, ignore_nbc_change=1):
        raise NotImplementedError("{!s}.getAOSPath(ignore_nbc_change=1)".format(self))

    @staticmethod
    def getAoSElement(self):
        logger.warning(
            "getAoSElement is deprecated, you should never need this", FutureWarning
        )
        return copy.deepcopy(self._element_structure)

    @staticmethod
    def getBackendInfo(parentCtx, index, homogeneousTime):  # Is this specific?
        raise NotImplementedError(
            "{!s}.getBackendInfo(parentCtx, index, homogeneousTime)".format(self)
        )

    def __init__(self, parent, name, structure_xml, base_path_in="element"):
        """Initialize IDSStructArray from XML specification

        Initializes in-memory an IDSStructArray. The XML should contain
        all direct descendants of the node. To avoid duplication,
        none of the XML structure is saved directly, so this transformation
        might be irreversible.

        Args:
          - parent: Parent structure. Can be anything, but at database write
                    time should be something with a path attribute
          - name: Name of the node itself. Will be used in path generation when
                  stored in DB
          - structure_xml: Object describing the structure of the IDS. Usually
                           an instance of `xml.etree.ElementTree.Element`
        """
        self._base_path = base_path_in
        self._convert_ids_types = False
        self._name = name
        self._parent = parent
        self._coordinates = {
            attr: structure_xml.attrib[attr]
            for attr in structure_xml.attrib
            if attr.startswith("coordinate")
        }
        # Save the converted structure_xml for later reference, and adding new
        # empty structures to the AoS
        self._element_structure = IDSStructure(self, name + "_el", structure_xml)
        # Do not try to convert ids_types by default.
        # As soon as a copy is made, set this to True
        self._element_structure._convert_ids_types = (
            False  # Enable converting after copy
        )
        # Do not store a reference to the parent. We will set this explicitly
        # each time a new instance is created, as all instances share the same
        # parent, this structure itself.
        self._element_structure._parent = None

        # Initialize with an 0-lenght list
        self.value = []

        self._convert_ids_types = True

    def __setattr__(self, key, value):
        object.__setattr__(self, key, value)

    def __getattr__(self, key):
        object.__getattribute__(self, key)

    def __getitem__(self, item):
        # value is a list, so the given item should be convertable to integer
        list_idx = int(item)
        return self.value[list_idx]

    def __setitem__(self, item, value):
        # value is a list, so the given item should be convertable to integer
        list_idx = int(item)
        if hasattr(self, "_convert_ids_types") and self._convert_ids_types:
            # Convert IDS type on set time. Never try this for hidden attributes!
            if list_idx in self.value:
                struct = self.value[list_idx]
                try:
                    struct.value = value
                except Exception as ee:
                    raise
        self.value[list_idx] = value

    def append(self, elt):
        """Append elements to the end of the array of structures.

        Parameters
        ----------
        """
        if not isinstance(elt, list):
            elements = [elt]
        else:
            elements = elt
        for e in elements:
            # Just blindly append for now
            # TODO: Maybe check if user is not trying to append weird elements
            self.value.append(e)

    def resize(self, nbelt, keep=False):
        """Resize an array of structures.

        Parameters
        ----------
        nbelt : int
            The number of elements for the targeted array of structure,
            which can be smaller or bigger than the size of the current
            array if it already exists.
        keep : bool, optional
            Specifies if the targeted array of structure should keep
            existing data in remaining elements after resizing it.
        """
        if not keep:
            self.value = []
        cur = len(self.value)
        if nbelt > cur:
            new_els = []
            for ii in range(nbelt - cur):
                new_el = copy.deepcopy(self._element_structure)
                new_el._parent = self
                new_el._convert_ids_types = True
                new_els.append(new_el)
            self.append(new_els)
        elif nbelt < cur:
            raise NotImplementedError("Making IDSStructArrays smaller")
            for i in range(nbelt, cur):
                self.value.pop()
        elif not keep:  # case nbelt = cur
            raise NotImplementedError("Overwriting IDSStructArray elements")
            self.append(
                [
                    process_charge_state__structArrayElement(self._base_path)
                    for i in range(nbelt)
                ]
            )

    def _getData(
        self, aosCtx, indexFrom, indexTo, homogeneousTime, nodePath, analyzeTime
    ):
        raise NotImplementedError(
            "{!s}._getData(aosCtx, indexFrom, indexTo, homogeneousTime, nodePath, analyzeTime)".format(
                self
            )
        )

    @loglevel
    def get(self, parentCtx, homogeneousTime):
        """Get data from UAL backend storage format and overwrite data in node

        Tries to dynamically build all needed information for the UAL.
        """
        timeBasePath = self.getTimeBasePath(homogeneousTime, 0)
        nodePath = self.getRelCTXPath(parentCtx)
        status, aosCtx, size = self._ull.ual_begin_arraystruct_action(
            parentCtx, nodePath, timeBasePath, 0
        )
        if status < 0:
            raise ALException(
                'ERROR: ual_begin_arraystruct_action failed for "process/products/element"',
                status,
            )

        if size < 1:
            return
        if aosCtx > 0:
            context_store[aosCtx] = (
                context_store[parentCtx] + "/" + nodePath + "/" + str(0)
            )
        self.resize(size)
        for i in range(size):
            self.value[i].get(aosCtx, homogeneousTime)
            self._ull.ual_iterate_over_arraystruct(aosCtx, 1)
            context_store.update(
                aosCtx, context_store[parentCtx] + "/" + nodePath + "/" + str(i + 1)
            )  # Update context

        if aosCtx > 0:
            context_store.pop(aosCtx)
            self._ull.ual_end_action(aosCtx)

    def getRelCTXPath(self, ctx):
        """ Get the path relative to given context from an absolute path"""
        if self.path.startswith(context_store[ctx]):
            rel_path = self.path[len(context_store[ctx]) + 1 :]
        else:
            raise Exception("Could not strip context from absolute path")
        return rel_path

    def put(self, parentCtx, homogeneousTime):
        """Put data into UAL backend storage format

        As all children _should_ support being put, just call `put` blindly.
        """
        timeBasePath = self.getTimeBasePath(homogeneousTime)
        # TODO: This might be to simple for array of array of structures
        nodePath = self.getRelCTXPath(parentCtx)
        status, aosCtx, size = self._ull.ual_begin_arraystruct_action(
            parentCtx, nodePath, timeBasePath, len(self.value)
        )
        if status != 0 or aosCtx < 0:
            raise ALException(
                'ERROR: ual_begin_arraystruct_action failed for "{!s}"'.format(
                    self._name
                ),
                status,
            )
        context_store[aosCtx] = context_store[parentCtx] + "/" + nodePath + "/" + str(0)

        for i in range(size):
            # This loops over the whole array
            dbg_str = " " * self.depth + "- [" + str(i) + "]"
            logger.debug("{:53.53s} put".format(dbg_str))
            self.value[i].put(aosCtx, homogeneousTime)
            status = self._ull.ual_iterate_over_arraystruct(aosCtx, 1)
            if status != 0:
                raise ALException(
                    'ERROR: ual_iterate_over_arraystruct failed for "{!s}"'.format(
                        self._name
                    ),
                    status,
                )
            context_store.update(
                aosCtx, context_store[parentCtx] + "/" + nodePath + "/" + str(i + 1)
            )  # Update context

        status = self._ull.ual_end_action(aosCtx)
        context_store.pop(aosCtx)
        if status != 0:
            raise ALException(
                'ERROR: ual_end_action failed for "{!s}"'.format(self._name), status
            )


class IDSToplevel(IDSStructure):
    """This is any IDS Structure which has ids_properties as child node

    At minium, one should fill ids_properties/homogeneous_time
    IF a quantity is filled, the coordinates of that quantity must be filled as well
    """

    @loglevel
    def readHomogeneous(self, occurrence):
        """Read the value of homogeneousTime.

        Returns:
            0: IDS_TIME_MODE_HETEROGENEOUS; Dynamic nodes may be asynchronous, their timebase is located as indicted in the "Coordinates" column of the documentation
            1: IDS_TIME_MODE_HOMOGENEOUS; All dynamic nodes are synchronous, their common timebase is the "time" node that is the child of the nearest parent IDS
            2: IDS_TIME_MODE_INDEPENDENT; No dynamic node is filled in the IDS (dynamic nodes _will_ be skipped by the Access Layer)
        """
        homogeneousTime = IDS_TIME_MODE_UNKNOWN
        if occurrence == 0:
            path = self._name
        else:
            path = self._name + "/" + str(occurrence)

        status, ctx = self._ull.ual_begin_global_action(self._idx, path, READ_OP)
        context_store[ctx] = context_store[self._idx] + "/" + path
        if status != 0:
            raise ALException(
                "Error calling ual_begin_global_action() in readHomogeneous() operation",
                status,
            )

        status, homogeneousTime = self._ull.ual_read_data(
            ctx, "ids_properties/homogeneous_time", "", INTEGER_DATA, 0
        )
        if status != 0:
            raise ALException("ERROR: homogeneous_time cannot be read.", status)
        status = self._ull.ual_end_action(ctx)
        context_store.pop(ctx)
        if status != 0:
            raise ALException(
                "Error calling ual_end_action() in readHomogeneous() operation", status
            )
        return homogeneousTime

    @loglevel
    def read_data_dictionary_version(self, occurrence):
        data_dictionary_version = ""
        path = self._name
        if occurrence != 0:
            path += "/" + str(occurrence)

        status, ctx = self._ull.ual_begin_global_action(self._idx, path, READ_OP)
        context_store[ctx] = context_store[self._idx] + "/" + path
        if status != 0:
            raise ALException(
                "Error calling ual_begin_global_action() in read_data_dictionary_version() operation",
                status,
            )

        status, data_dictionary_version = self._ull.ual_read_data_string(
            ctx, "ids_properties/version_put/data_dictionary", "", CHAR_DATA, 1
        )
        if status != 0:
            raise ALException("ERROR: data_dictionary_version cannot be read.", status)
        status = self._ull.ual_end_action(ctx)
        context_store.pop(ctx)
        if status != 0:
            raise ALException(
                "Error calling ual_end_action() in read_data_dictionary_version() operation",
                status,
            )
        return data_dictionary_version

    @loglevel
    def get(self, occurrence=0, **kwargs):
        """Get data from UAL backend storage format and overwrite data in node

        Tries to dynamically build all needed information for the UAL. As this
        is the root node, it is simple to construct UAL paths and contexts at
        this level. Should have an open database.
        """
        path = None
        if occurrence == 0:
            path = self._name
        else:
            path = self._name + "/" + str(occurrence)

        homogeneousTime = self.readHomogeneous(occurrence)
        if homogeneousTime == IDS_TIME_MODE_UNKNOWN:
            logger.error(
                "Unknown time mode {!s}, stop getting of {!s}".format(
                    homogeneousTime, self._name
                )
            )
            return
        data_dictionary_version = self.read_data_dictionary_version(occurrence)

        # TODO: Do not use global context
        status, ctx = self._ull.ual_begin_global_action(self._idx, path, READ_OP)
        if status != 0:
            raise ALException(
                "Error calling ual_begin_global_action() for {!s}".format(self._name),
                status,
            )
        context_store[ctx] = context_store[self._idx] + path

        logger.debug("{:53.53s} get".format(self._name))
        super().get(ctx, homogeneousTime, **kwargs)

        status = self._ull.ual_end_action(ctx)
        context_store.pop(ctx)
        if status != 0:
            raise ALException(
                "Error calling ual_end_action() for {!s}".format(self._name), status
            )

    @loglevel
    def deleteData(self, occurrence=0):
        """Delete UAL backend storage data

        Tries to dynamically build all needed information for the UAL. As this
        is the root node, it is simple to construct UAL paths and contexts at
        this level. Should have an open database.
        """
        if not np.issubdtype(type(occurrence), np.integer):
            raise ValuError("Occurrence should be an integer")

        rel_path = self.getRelCTXPath(self._idx)
        if occurrence != 0:
            rel_path += "/" + str(occurrence)

        status, ctx = self._ull.ual_begin_global_action(self._idx, rel_path, WRITE_OP)
        context_store[ctx] = context_store[self._idx] + rel_path
        if status < 0:
            raise ALException(
                'ERROR: ual_begin_global_action failed for "{!s}"'.format(rel_path),
                status,
            )

        for child_name in self._children:
            child = getattr(self, child_name)
            if isinstance(child, (IDSStructArray, IDSPrimitive)):
                status = self._ull.ual_delete_data(ctx, child_name)
                if status != 0:
                    raise ALException(
                        'ERROR: ual_delete_data failed for "{!s}". Status code {!s}'.format(
                            rel_path + "/" + child_name
                        ),
                        status,
                    )
            else:
                status = child.delete(ctx)
                if status != 0:
                    raise ALException(
                        'ERROR: delete failed for "{!s}". Status code {!s}'.format(
                            rel_path + "/" + child_name
                        ),
                        status,
                    )
        status = self._ull.ual_end_action(ctx)
        context_store.pop(ctx)
        if status < 0:
            raise ALException(
                'ERROR: ual_end_action failed for "{!s}"'.format(rel_path), status
            )
        return 0

    @loglevel
    def to_ualstore(self, ual_data_store, path=None, occurrence=0):
        """Put data into UAL backend storage format

        As all children _should_ support being put, just call `put` blindly.

        Tries to dynamically build all needed information for the UAL. As this
        is the root node, it is simple to construct UAL paths and contexts at
        this level. Should have an open database.
        """
        if path is not None:
            raise NotImplementedError("Explicit paths, implicitly handled by structure")

        path = self.path
        if occurrence != 0:
            path += "/" + str(occurrence)

        # Determine the time_mode.
        homogeneousTime = self.ids_properties.homogeneous_time.value
        if homogeneousTime == IDS_TIME_MODE_UNKNOWN:
            logger.warning(
                "IDS {!s} is found to be EMPTY (homogeneous_time undefined). PUT quits with no action."
            )
            return
        if homogeneousTime not in IDS_TIME_MODES:
            raise ALException(
                "ERROR: ids_properties.homogeneous_time should be set to IDS_TIME_MODE_HETEROGENEOUS, IDS_TIME_MODE_HOMOGENEOUS or IDS_TIME_MODE_INDEPENDENT."
            )
        if homogeneousTime == IDS_TIME_MODE_HOMOGENEOUS and len(self.time.value) == 0:
            raise ALException(
                "ERROR: the IDS%time vector of an homogeneous_time IDS must have a non-zero length."
            )

        # Delete the data in the store
        # TODO: handle mode correctly!
        self.deleteData(occurrence)

        # Begin a write action
        status, ctx = self._ull.ual_begin_global_action(self._idx, path, WRITE_OP)
        if status != 0:
            raise ALException(
                "Error calling ual_begin_global_action() for {!s}".format(
                    self._name, status
                )
            )

        context_store[ctx] = path
        for child_name in self._children:
            child = getattr(self, child_name)
            dbg_str = " " * self.depth + "- " + child_name
            if not isinstance(child, IDSPrimitive):
                logger.debug("{:53.53s} put".format(dbg_str))
            child.put(ctx, homogeneousTime)

        context_store.pop(ctx)
        status = self._ull.ual_end_action(ctx)
        if status != 0:
            raise ALException(
                "Error calling ual_end_action() for {!s}".format(self._name), status
            )

    def setExpIdx(self, idx):
        logger.warning(
            "setExpIdx is deprecated, call self.setPulseCtx instead", FutureWarning
        )
        self.setPulseCtx(idx)

    @loglevel
    def put(self, occurrence=0, data_store=None):
        if data_store is None:
            data_store = self._data_store
        self.to_ualstore(data_store, path=None, occurrence=occurrence)

    @property
    def _data_store(self):
        return self._parent._data_store

    @property
    def _idx(self):
        return self._data_store._idx

    @classmethod
    def getMaxOccurrences(self):
        raise NotImplementedError("{!s}.getMaxOccurrences()".format(self))
        return cls._MAX_OCCURRENCES

    def initIDS(self):
        raise NotImplementedError("{!s}.initIDS()".format(self))

    def partialGet(self, dataPath, occurrence=0):
        raise NotImplementedError(
            "{!s}.partialGet(dataPath, occurrence=0)".format(self)
        )

    def getField(self, dataPath, occurrence=0):
        raise NotImplementedError("{!s}.getField(dataPath, occurrence=0)".format(self))

    def _getFromPath(self, dataPath, occurrence, analyzeTime, data_store=None):
        # Retrieve partial IDS data without reading the full database content
        raise NotImplementedError(
            "{!s}.getField(dataPath, occurrence, analyzeTime, data_store=None)".format(
                self
            )
        )



# FileManager and DummyFileManager
# are taken from xarray c07160dd2d627a021e58515cbd7753c11fb56d94
#
# This code can be modified if you explicitly note your
# changes. However, it would be better to wrap these
# classes instead
#
#
# Copyright 2014-2020, xarray Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib


class FileManager:
    """Manager for acquiring and closing a file object.

    Use FileManager subclasses (CachingFileManager in particular) on backend
    storage classes to automatically handle issues related to keeping track of
    many open files and transferring them between multiple processes.
    """

    def acquire(self, needs_lock=True):
        """Acquire the file object from this manager."""
        raise NotImplementedError()

    def acquire_context(self, needs_lock=True):
        """Context manager for acquiring a file. Yields a file object.

        The context manager unwinds any actions taken as part of acquisition
        (i.e., removes it from any cache) if an exception is raised from the
        context. It *does not* automatically close the file.
        """
        raise NotImplementedError()

    def close(self, needs_lock=True):
        """Close the file object associated with this manager, if needed."""
        raise NotImplementedError()


class DummyFileManager(FileManager):
    """FileManager that simply wraps an open file in the FileManager interface."""

    def __init__(self, value):
        self._value = value

    def acquire(self, needs_lock=True):
        del needs_lock  # ignored
        return self._value

    @contextlib.contextmanager
    def acquire_context(self, needs_lock=True):
        del needs_lock
        yield self._value

    def close(self, needs_lock=True):
        del needs_lock  # ignored
        self._value.close()



# NdimSizeLenMixin and NDArrayMixin
# are taken from xarray c07160dd2d627a021e58515cbd7753c11fb56d94
#
# This code can be modified if you explicitly note your
# changes. However, it would be better to wrap these
# classes instead
#
#
# Copyright 2014-2020, xarray Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    AbstractSet,
    Any,
    Callable,
    Collection,
    Container,
    Dict,
    Hashable,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    MutableSet,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import numpy as np


class NdimSizeLenMixin:
    """Mixin class that extends a class that defines a ``shape`` property to
    one that also defines ``ndim``, ``size`` and ``__len__``.
    """

    __slots__ = ()

    @property
    def ndim(self: Any) -> int:
        return len(self.shape)

    @property
    def size(self: Any) -> int:
        # cast to int so that shape = () gives size = 1
        return int(np.prod(self.shape))

    def __len__(self: Any) -> int:
        try:
            return self.shape[0]
        except IndexError:
            raise TypeError("len() of unsized object")


class NDArrayMixin(NdimSizeLenMixin):
    """Mixin class for making wrappers of N-dimensional arrays that conform to
    the ndarray interface required for the data argument to Variable objects.

    A subclass should set the `array` property and override one or more of
    `dtype`, `shape` and `__getitem__`.
    """

    __slots__ = ()

    @property
    def dtype(self: Any) -> np.dtype:
        return self.array.dtype

    @property
    def shape(self: Any) -> Tuple[int]:
        return self.array.shape

    def __getitem__(self: Any, key):
        return self.array[key]

    def __repr__(self: Any) -> str:
        return "{}(array={!r})".format(type(self).__name__, self.array)



# This file is part of IMASPy.
#
# IMASPy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IMASPy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IMASPy.  If not, see <https://www.gnu.org/licenses/>.
import sys
import os
import logging

from IPython import embed
import numpy
import numpy as np

root_logger = logging.getLogger("imaspy")
logger = root_logger
logger.setLevel(logging.WARNING)

if __name__ == "__main__":
    # from imaspy.ids_classes import *
    # MVP, get a slice
    # So for example be able to run
    # import imaspy as imas
    # input = imas.ids(shot,run_in)
    # input.open_env(input_user_or_path,input_database,’3’)
    # input.equilibrium.getSlice(time_slice, 1)

    # Existing workflow example
    shot = 130012
    # input_user_or_path = 'public'
    input_database = "iter"
    time_slice = 200
    occurence = 0

    # Test example
    TEST_SHOT = 1
    TEST_RUN = 2
    backend = 3
    BACKEND_ID_0 = 10
    NO_BACKEND = BACKEND_ID_0
    ASCII_BACKEND = BACKEND_ID_0 + 1
    MDSPLUS_BACKEND = BACKEND_ID_0 + 2
    HDF5_BACKEND = BACKEND_ID_0 + 3
    MEMORY_BACKEND = BACKEND_ID_0 + 4
    UDA_BACKEND = BACKEND_ID_0 + 5
    # backend = MEMORY_BACKEND
    backend = MDSPLUS_BACKEND
    ACCESS_PULSE_0 = 40
    FORCE_CREATE_PULSE = ACCESS_PULSE_0 + 3
    input_user_or_path = "vandepk"

    # Old API
    # import matplotlib.pyplot as plt
    style = "old"
    style = "new"
    if style == "old":
        # import imas_3_28_1_ual_4_7_2_dev38 as imas
        run_in = 0
        try:
            # Try loading Karels special library
            import imas_3_28_1_ual_4_7_2_dev36 as imas
        except ImportError:
            # Load the regular IMAS library
            import imas

        imas_entry = imas.ids(shot, run_in)
        idx = imas_entry.create_env_backend(
            input_user_or_path, input_database, "3", backend
        )
    else:
        # New API
        run_in = 1
        # 4.4.1 Initializing an empty IDS
        import imaspy as imas
        from pathlib import Path

        imas.ids = imas.ids_classes.IDSRoot  # Enable old-style syntax

        IMAS_PREFIX = os.getenv("IMAS_PREFIX")
        if not IMAS_PREFIX or not os.path.isdir(IMAS_PREFIX):
            logger.warning(
                "IMAS_PREFIX is unset or is not a directory. Points to {!s}. Fall back to 'karel paths'!".format(
                    IMAS_PREFIX
                )
            )
            idsdef_dir = Path(os.path.dirname(__file__)) / Path(
                "../../../../working/imas-data-dictionary/"
            )
            idsdef = Path(idsdef_dir) / Path("IDSDef.xml")
        else:
            imas_prefix = Path(os.getenv("IMAS_PREFIX"))
            imas_include = imas_prefix.joinpath("include")
            idsdef = imas_include.joinpath("IDSDef.xml")
        imas_entry = imas.ids(shot, run_in, xml_path=idsdef, verbosity=2)
        idx = imas_entry.create_env_backend(
            input_user_or_path, input_database, "3", backend
        )
    # This should then support accessing an IDS using idsVar.<IDSname>.<IDSfield>

    # 5.5 Opening an existing Data Entry
    # imas_entry.open_env(input_user_or_path, input_database, '3')

    # 5.6 Opening a remote Data Entry using the UDA backend
    # imas_entry = imas.ids(54178, 0)
    # imas_entry.open_public('WEST')

    # 5.7 Closing an opened Data Entry
    # [retstatus] = imas_entry.close()

    # 4.4.2 Copying a variable to a new IDS. Not deepcopy! Proposed API:
    # idsVar2.equilibrium.copyValues(idsVar)

    # 4.4.3 Deallocation
    # Using python build-in management, or del var

    # 5.4 Creating a new data entry
    # Put mandatory top-level entry
    ids = imas_entry.equilibrium
    # ids.ids_properties.homogeneous_time = 2
    time = imas._libs.imasdef.IDS_TIME_MODE_HOMOGENEOUS
    ids.ids_properties.homogeneous_time = time

    ids.put()
    print(
        "Original ids.ids_properties.homogeneous_time:",
        ids.ids_properties.homogeneous_time,
    )
    ids.ids_properties.homogeneous_time = 9999
    print(
        "Changed ids.ids_properties.homogeneous_time:",
        ids.ids_properties.homogeneous_time,
    )
    ids.get()
    print(
        "After re-get ids.ids_properties.homogeneous_time:",
        ids.ids_properties.homogeneous_time,
    )

    # 5.8 Putting an IDS
    # 5.10 Deleting an IDS # ids.put() deletes the IDS, so calls deleteData(self, occurence)
    # 5.11 Getting an IDS # Easiest to show the put works by showing a get!

    # Creating a nested data entry
    print(
        "Pre put ids.ids_properties.version_put.access_layer:",
        ids.ids_properties.version_put.access_layer,
    )
    ids.ids_properties.version_put.access_layer = "imaspy"
    ids.put()
    print(
        "Post put ids.ids_properties.version_put.access_layer:",
        ids.ids_properties.version_put.access_layer,
    )
    ids.get()
    print(
        "Post get ids.ids_properties.version_put.access_layer:",
        ids.ids_properties.version_put.access_layer,
    )

    # Deeper fields
    # Set time vector
    print("Pre put ids.time:", ids.time)
    ids.time = np.array([0.1, 0.2, 0.3])
    ids.time = ids.time + 2
    ids.put()
    print("Post put ids.time:", ids.time)
    ids.time = [12345678]  # Scamble
    ids.get()
    print("Post get ids.time:", ids.time)

    # And an array
    print("Pre put ids.vacuum_toroidal_field.b0:", ids.vacuum_toroidal_field.b0)
    ids.vacuum_toroidal_field.b0 = np.array([1, 2, 3])
    ids.put()
    print("Post put ids.vacuum_toroidal_field.b0:", ids.vacuum_toroidal_field.b0)
    # Scramble
    ids.vacuum_toroidal_field.b0 = [-98767890]
    ids.get()
    print("Post get ids.vacuum_toroidal_field.b0:", ids.vacuum_toroidal_field.b0)

    # We can do fancy math even!
    print("ids.vacuum_toroidal_field.b0 + 2:", ids.vacuum_toroidal_field.b0 + 2)

    ## Set beyond a struct array
    print("Pre put ids.time_slice:", ids.time_slice)
    ids.time_slice.resize(1)
    ids.time_slice[0].profiles_1d.psi = np.array([0, 0.5, 1.5])
    ids.time_slice[0].time = 0.1
    ids.put()
    print("Post put ids.time_slice[0].time:", ids.time_slice[0].time)
    print(
        "Post put ids.time_slice[0].profiles_1d.psi:", ids.time_slice[0].profiles_1d.psi
    )
    # Scramble
    ids.time_slice[0].time = [-12344567]
    ids.time_slice[0].profiles_1d.psi = [-98767890]
    ids.get()
    print("Post get ids.time_slice[0].time:", ids.time_slice[0].time)
    print(
        "Post get ids.time_slice[0].profiles_1d.psi:", ids.time_slice[0].profiles_1d.psi
    )

    ids.time_slice[0].profiles_2d.resize(1)
    ids.time_slice[0].profiles_2d[0].grid_type.name = "I am grid"
    ids.put()
    print(
        "Post put ids.time_slice[0].profiles_2d[0].grid_type.name:",
        ids.time_slice[0].profiles_2d[0].grid_type.name,
    )
    ## Scramble
    ids.time_slice[0].profiles_2d[0].grid_type.name = "stuff"

    ids.get()
    print(
        "Post get ids.time_slice[0].profiles_2d[0].grid_type.name:",
        ids.time_slice[0].profiles_2d[0].grid_type.name,
    )

    # time_slice.coordinate_system.grid.volume_element 2D_FLT
    # time_slice.coordinate_system.tensor_covariant 4D_FLT

    # 5.9 Putting an IDS progressively in a time loop
    # Expect the IDS variable to contain a single slice in any non-empty dynamic nodes
    # So, start with a single .put, then progressive put_slice IN STRICT TIME ORDER

    # 5.12 Getting a slice
    # imas_entry.pf_active.getSlice(0.1, 2)

    # 5.13 Getting units of a note
    # Slash-separated path, ignore arrays, as all units in array are the same
    # units = get_units(ids, node_path)

    # 5.14 Getting a subset of an IDS
    # Smart datapath, so path/to/array(idx)/field
    # Smart datapath, so path/to/array(x:y)/field
    # partialGet

    # 5.15 Selecting the backend
    # MDSPLUS, MEMORY, UDA, ASCII
    # MAGIC, ignore for now?

    print("Done!")



# A minimal testcase loading an IDS file and checking that the structure built is ok

import logging

import imaspy

root_logger = logging.getLogger("imaspy")
logger = root_logger
logger.setLevel(logging.DEBUG)


def test_load_minimal_types():
    ids = imaspy.ids_classes.IDSRoot(
        0, 0, xml_path="assets/IDS_minimal_types.xml", verbosity=2
    )  # Create a empty IDSs

    # Check if the datatypes are loaded correctly
    assert ids.minimal.a.data_type == "FLT_0D"
    assert ids.minimal.ids_properties.comment.data_type == "STR_0D"



# This file is part of IMASPy.
#
# IMASPy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IMASPy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IMASPy.  If not, see <https://www.gnu.org/licenses/>.

import pkg_resources
from distutils.version import LooseVersion as V
try:
    __version__ = pkg_resources.get_distribution("imaspy").version
except Exception:
    # Try local wrongly install copy
    try:
        from version import __version__
    except Exception:
        # Local copy or not installed with setuptools.
        # Disable minimum version checks on downstream libraries.
        __version__ = "0.0.0"

import logging
root_logger = logging.getLogger("imaspy")
if V(__version__) < V("0.2"):
    root_logger.critical("You are using the FOSS version of IMASPy. Please use the ITER version"
                         " instead! See `https://git.iter.org/projects/IMAS/repos/imaspy`")
from imaspy import (
    setup_logging,
    imas_ual_env_parsing,
    ids_classes,
)

from imaspy.backends import (
    file_manager,
    xarray_core_indexing,
    xarray_core_utils,
    common,
    ual,
)



# This file is part of IMASPy.
#
# IMASPy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IMASPy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IMASPy.  If not, see <https://www.gnu.org/licenses/>.
import logging
import time

TRACE_LEVEL_NUM = 5
logging.addLevelName(TRACE_LEVEL_NUM, "TRACE")


def trace(self, message, *args, **kws):
    if self.isEnabledFor(TRACE_LEVEL_NUM):
        # Yes, logger takes its '*args' as 'args'.
        self._log(TRACE_LEVEL_NUM, message, args, **kws)
        logging.Logger.trace = trace


logging.Logger.trace = trace
logging.TRACE = TRACE_LEVEL_NUM


class PrettyFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""

    light_grey = "\x1b[38;5;251m"
    yellow = "\x1b[33m"
    red = "\x1b[31m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    formatstr = "%(asctime)s [%(levelname)s] %(message)s @%(filename)s:%(lineno)d"
    default_time_format = "%H:%M:%S"

    FORMATS = {
        logging.TRACE: light_grey + formatstr + reset,
        logging.DEBUG: light_grey + formatstr + reset,
        logging.INFO: formatstr,
        logging.WARNING: yellow + formatstr + reset,
        logging.ERROR: red + formatstr + reset,
        logging.CRITICAL: bold_red + formatstr + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        formatter.formatTime = self.formatTime
        return formatter.format(record)

    def formatTime(self, record, datefmt=None, print_ms=False):
        # pylint: disable=arguments-differ # Adds print_ms kwarg
        ct = self.converter(record.created)
        if datefmt:
            ss = time.strftime(datefmt, ct)
        else:
            tt = time.strftime(self.default_time_format, ct)
            if print_ms:
                ss = self.default_msec_format % (tt, record.msecs)
            else:
                ss = tt
        return ss


def test_messages():
    """ Print out a message on each logging level """
    logger = logging.getLogger("imaspy.testlogger")
    logger.setLevel(logging.TRACE)
    logger.trace("Trace message")
    logger.debug("Debug message")
    logger.info("Info message")
    logger.warning("Warning message")
    logger.error("Error message")
    logger.critical("Critical message")


def connect_formatter(logger):
    """ Connect general formatter to given logger """
    ch = logging.StreamHandler()
    ch.setLevel(logging.TRACE)
    ch.setFormatter(PrettyFormatter())
    logger.addHandler(ch)


# Log to console by default, and output it all
root_logger = logging.getLogger("imaspy")
connect_formatter(root_logger)

if __name__ == "__main__":
    test_messages()



# -------------------------------------------------
# PYTHON WRAPPER TO CALL PION (AND OTIONALLY NEMO)
# -------------------------------------------------

import copy

# NEEDED MODULES
import os
import sys

from IPython import embed  # Convinience

# INPUT/OUTPUT CONFIGURATION
shot = 44
run_in = 33
run_out = 133
input_user_or_path = "public"
input_database = "ITER_MD"
output_user_or_path = os.getenv("USER")
output_database = input_database

# DISPLAY SIMULATION INFORMATION
print("---------------------------------")
print("shot                = ", shot)
print("run_in              = ", run_in)
print("run_out             = ", run_out)
print("input_user_or_path  = ", input_user_or_path)
print("input_database      = ", input_database)
print("output_user_or_path = ", output_user_or_path)
print("output_database     = ", output_database)
print("---------------------------------")

# IF THE OUTPUT DATABASE DOES NOT EXIST: CREATE IT
if output_user_or_path == os.getenv("USER"):
    output_folder = os.getenv("HOME") + "/public/imasdb/" + output_database + "/3/0"
else:
    output_folder = output_user_or_path + "/" + output_database + "/3/0"
if os.path.isdir(output_folder) == False:
    print("-- Create local database for output file " + output_folder, file=sys.stdout)
    os.makedirs(output_folder)

# IMAS DATA VERSION
version = os.getenv("IMAS_VERSION")[0]

# OPEN INPUT DATAFILE
style = "old"
style = "new"
if style == "old":
    import imas

    print("=> Open input datafile")
    input = imas.DBEntry(
        imas.imasdef.MDSPLUS_BACKEND, input_database, shot, run_in, input_user_or_path
    )
    input.open()

    # READ NBI IDS FROM INPUT DATAFILE
    print("=> Read nbi IDS")
    input_nbi = input.get("nbi")

    # COPY THE NBI IDS (IN VIEW OF MODIFYING IT)
    output_nbi = input_nbi  # I don't think this actually makes a copy!!!

    # CREATE OUTPUT DATAFILE
    print("=> Create output datafile")
    output = imas.DBEntry(
        imas.imasdef.MDSPLUS_BACKEND,
        output_database,
        shot,
        run_out,
        output_user_or_path,
    )
    output.create()

    # WRITE NBI IDS IN OUTPUT DATAFILE
    print("=> Write nbi IDS")
    output.put(output_nbi)

    # CLOSE FILES
    input.close()
    output.close()
    print("Done.")
else:
    import imaspy as imas
    from imas.imasdef import MDSPLUS_BACKEND
    from pathlib import Path

    imas.ids = imas.ids_classes.IDSRoot  # Enable old-style syntax
    imas_prefix = Path(os.getenv("IMAS_PREFIX"))
    imas_include = imas_prefix.joinpath("xml")
    idsdef = imas_include.joinpath("IDSDef.xml")
    input = imas.ids(shot, run_in, xml_path=idsdef, verbosity=2)  # Create a empty IDSs
    input.open_env_backend(
        input_user_or_path, input_database, "3", MDSPLUS_BACKEND
    )  # Start a new 'pulse action', linking IDSs to input database

    # READ NBI IDS FROM INPUT DATAFILE
    input_nbi = input.nbi  # Make a convinient name for NBI IDS
    input_nbi.get()  # Fill empty nbi structure from database

    output_nbi = copy.deepcopy(
        input_nbi
    )  # This really copies the IDS. Not for old API though..

    ## Do a hacky thing to link to a new file now, should mirror the Python HLI API I guess
    output = output_nbi._parent
    output.create_env_backend(
        output_user_or_path, output_database, "3", MDSPLUS_BACKEND
    )  # Start a new 'pulse acting', linking IDSs to output database

    # embed() # Tip: Use IPython's embed() or ipdb for autocomplete, colors, etc.
    unit = output_nbi["unit"][1]
    unit["name"] = "banana"

    # WRITE NBI IDS IN OUTPUT DATAFILE
    output_nbi.put()

    # CLOSE FILES
    input.close()
    output.close()
    print("Done.")



import pytest
import types
import imaspy


def test_access_layer_version():
    def test_access_layer_version_version():
        assert isinstance(imaspy.__version__, str)

    @pytest.mark.xfail
    def test_expected_fail():
        assert False

    @pytest.mark.xfail
    def test_unexpected_pass():
        assert True



# This file is part of IMASPy.
#
# IMASPy is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IMASPy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IMASPy.  If not, see <https://www.gnu.org/licenses/>.
def parse_UAL_version_string(string):
    """Parses a UAL_VERSION string

    Parses the UAL_VERSION to IMASPy format. By convention, the
    UAL_VERSION string is generated by git-describe and edited.
    In our case for patch versions 'symver', e.g. '4.8.2'. For
    development versions it is 'symver-steps_from_symver-gcommit', e.g.
    '4.8.2-2-g1fffb6bf'

    Args:
      - string: The UAL_VERSION string.

    Returns:
      - ual_symver: The symversion of the passed UAL_VERSION
      - steps_from_version: Amount of steps the development version is from symver
      - ual_commit: The exact commit of the passed string. Symver for release version.

    """

    if "-" in string:
        ual_symver, micropatch = string.split("-", 1)
        steps_from_version, commitspec = micropatch.split("-", 2)
        ual_commit = commitspec[1:]
    else:
        ual_symver = string
        steps_from_version = "0"
        ual_commit = ual_symver
    return ual_symver, steps_from_version, ual_commit


def sanitise_UAL_symver(ual_symver):
    """Sanitizes a ual_symver for use in packages"""
    return ual_symver.replace(".", "_")


def build_UAL_package_name(sanitised_UAL_symver, ual_commit):
    """Build the package name for UAL Python packages"""
    return "imas._ual_lowlevel"

    if "." in ual_commit:
        # Release package
        return "ual_{!s}._ual_lowlevel".format(sanitised_UAL_symver)
    else:
        # 'dev' or 'in-between' package
        return "ual_{!s}_{!s}._ual_lowlevel".format(sanitised_UAL_symver, ual_commit)
